/*
 *  Lightweight At Daemon
 *
 *  Compile with:
 *    gcc -s -Wall -Wstrict-prototypes -Os atd.c -o atd
 *
 *  Copyright (C) 1996, Paul Gortmaker
 *  Copyright (C) 2001, Russell Nelson
 *  Copyright (C) 2002-2004, Nils Faerber <nils@handhelds.org>
 *
 *  Released under the GNU General Public License, version 2,
 *  included herein by reference.
 *
 *  ChangeLog:
 *  2006-12-20  YANG Lin  <lin.yang@orange-ftgroup.com>
 *  Modify it to fit for the reference device (C8000).         
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <linux/rtc.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/select.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <limits.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <dirent.h>
#include <pwd.h>
#include <signal.h>

#include <calenabler/cal.h>
#define DEBUG
#define c8000


#define PCF_SET_ALARM 0x300
#define PCF_GET_ALARM 0x400
#define PCF_ENABLE_ALARM 0x500
#define PCF_DISABLE_ALARM 0x600



/* globals... */
int scan_i = 0;
DIR *scan_dir = NULL;
char *dirs[] = { "/var/spool/at", /* "/var/spool/cron", */ NULL };

typedef struct
{
  int year;
  int mon;
  int day;
  int wkday;
  int hour;
  int minute;
  int second;
} alarm_datetime_t;

void alarmSignalHandler (int sig);


int
compare_rtc_to_tm (struct rtc_time *rtc, struct tm *tm)
{
  int i;

  i = rtc->tm_year - tm->tm_year;
  if (!i)
    i = rtc->tm_mon - tm->tm_mon;
  if (!i)
    i = rtc->tm_mday - tm->tm_mday;
  if (!i)
    i = rtc->tm_hour - tm->tm_hour;
  if (!i)
    i = rtc->tm_min - tm->tm_min;
  if (!i)
    i = rtc->tm_sec - tm->tm_sec;
  if (!i)
    return 0;
  if (i > 0)
    return 1;
  return -1;
}



void
scan_from_top (void)
{
  if (scan_dir)
  {
    closedir (scan_dir);
    scan_dir = NULL;
    scan_i = 0;
  }
}


struct dirent *
scan (void)
{
  struct dirent *dirent;

  if (!scan_dir)
  {
    if (!dirs[scan_i])
    {
      scan_i = 0;
      return NULL;
    }
    else
      scan_dir = opendir (dirs[scan_i++]);
  }
  dirent = readdir (scan_dir);
  if (dirent)
    return dirent;
  closedir (scan_dir);
  scan_dir = NULL;

  return scan ();
}
unsigned long
nameok (const char *name)
{
  char *middle;
  unsigned long this;

  printf ("%s:name=%s\n", __FUNCTION__, name);
  /* check for timestamp.pid */
  this = strtol (name, &middle, 10);
  if (*middle != '.')
    return 0;

  return this;
}
unsigned long
setNearestAlarm ()
{
  struct dirent *dirent;
  unsigned long this, next, now;

  now = time (NULL);
  printf ("%s:now=%d\n", __FUNCTION__, now);
  scan_from_top ();
  next = ULONG_MAX;
  while ((dirent = scan ()))
  {
    this = nameok (dirent->d_name);
    printf ("%s:this=%d\n", __FUNCTION__, this);

    /* schedule jobs that we haven't just run */
    if (this)
      if (this > now + 1)
        if (this < next)
          next = this;
  }                             /* end while */
  if (next == ULONG_MAX)
    next = 0;
  return next;
}

void
waitfor (time_t t)
{
  int rtcfd, tfd, retval = 0;
  unsigned long data;
  struct rtc_time rtc_tm;
  time_t now, then;
  struct tm *tm;
  struct timeval tv;
  int nfds;
  fd_set afds;
  alarm_datetime_t ad;


#ifdef DEBUG
  printf ("waitfor %ld\n", t);
#endif
  rtcfd = open ("/dev/c8k_alarm", O_RDWR);
  if (rtcfd == -1)
  {
    perror ("/dev/c8k_alarm");
    exit (errno);
  }

  /* Read the RTC time/date */
  tfd = open ("trigger", O_RDWR | O_NONBLOCK);
  if (tfd == -1)
  {
    perror ("trigger");
    exit (errno);
  }

  /* Set the RTC time/date */
  now = time (NULL);
  tm = gmtime (&now);
  rtc_tm.tm_mday = tm->tm_mday;
  rtc_tm.tm_mon = tm->tm_mon;
  rtc_tm.tm_year = tm->tm_year;
  rtc_tm.tm_hour = tm->tm_hour;
  rtc_tm.tm_min = tm->tm_min;
  rtc_tm.tm_sec = tm->tm_sec;


  tm = gmtime (&t);

  /*If the time to set is before now, then return */
  if (t && compare_rtc_to_tm (&rtc_tm, tm) >= 0)
  {
    close (rtcfd);
    close (tfd);
    return;
  }

  if (t)
  {

#ifdef DEBUG
    printf ("Alarm date/time is %d-%d-%d, %02d:%02d:%02d.\n",
            tm->tm_mday, tm->tm_mon + 1, tm->tm_year + 1900,
            tm->tm_hour, tm->tm_min, tm->tm_sec);
#endif

    /* set the alarm */
    ad.year = tm->tm_year + 1900;
    ad.mon = tm->tm_mon + 1;
    ad.day = tm->tm_mday;
    ad.wkday = tm->tm_wday;
    ad.hour = tm->tm_hour;
    ad.minute = tm->tm_min;
    ad.second = tm->tm_sec;
    retval = ioctl (rtcfd, PCF_SET_ALARM, &ad);
    if (retval == -1)
    {
      perror ("ioctl");
      exit (errno);
    }
#ifdef DEBUG
    printf ("Set CECAlarm to PCF:%d-%d-%d, %02d:%02d:%02d.\n",
            ad.day, ad.mon, ad.year, ad.hour, ad.minute, ad.second);
#endif
    ad.year = 0;
    ad.mon = 0;
    ad.day = 0;
    /* Read the current alarm settings */
    retval = ioctl (rtcfd, PCF_GET_ALARM, &ad);
    if (retval == -1)
    {
      perror ("ioctl");
      exit (errno);
    }

#ifdef DEBUG
    printf ("Alarm date/time now in PCF: %d-%d-%d, %02d:%02d:%02d.\n",
            ad.day, ad.mon, ad.year, ad.hour, ad.minute, ad.second);
#endif

    /* Enable alarm interrupts */
    retval = ioctl (rtcfd, PCF_ENABLE_ALARM, NULL);
    if (retval == -1)
    {
      perror ("ioctl");
      exit (errno);
    }

    close (rtcfd);
  }

#ifdef DEBUG
  fflush (stdout);
  printf ("wait for adding alarm\n");
#endif
  FD_ZERO (&afds);
  FD_SET (tfd, &afds);
  nfds = tfd + 1;
  tv.tv_sec = 10 * 60;
  tv.tv_usec = 0;
  then = time (NULL);
/*

  if (select (nfds, &afds, (fd_set *) 0, (fd_set *) 0, &tv) < 0) 
    {
      if (errno != EINTR)
        perror ("select");
      exit (errno);
    }
*/
  int ret;
  ret = select (nfds, &afds, (fd_set *) 0, (fd_set *) 0, &tv);
  printf ("%s:ret = %d\n", __FUNCTION__, ret);
  now = time (NULL);
#ifdef DEBUG
  printf ("While we were sleeping, %d seconds elapsed, but %d alarms passed\n",
          (int) (now - then), (int) (tv.tv_sec));
#endif
  if (ret > 0)
    if (FD_ISSET (tfd, &afds))
    {
      printf ("tfd is ready\n");
      retval = read (tfd, &data, 1);
#ifdef DEBUG
      printf ("%s:data=%d\n", __FUNCTION__, data);
#endif
      if (retval == -1)
      {
        perror ("read");
        exit (errno);
      }
    }
  close (tfd);
}


int
runjob (char *fn)
{
  pid_t pid;
//unsigned int status;
  struct stat sbuf;
  pid = fork ();
  if (pid == -1)
    return -1;
  if (pid == 0)
  {
    if (getuid () == 0 && geteuid () == 0)
    {
      struct passwd *userpwd = NULL;
#ifdef DEBUG
      printf ("I am root so I am going to suid() me to ");
#endif
      stat (fn, &sbuf);
      if (setgid (sbuf.st_gid) != 0)
        perror ("setgid");
      if (setuid (sbuf.st_uid) != 0)
        perror ("setuid");
#ifdef DEBUG
      printf ("uid %d and gid %d\n", sbuf.st_uid, sbuf.st_gid);
#endif
      userpwd = getpwuid (sbuf.st_uid);
      if (userpwd != NULL)
      {
        setenv ("USER", userpwd->pw_name, 1);
        setenv ("LOGNAME", userpwd->pw_name, 1);
        setenv ("HOME", userpwd->pw_dir, 1);
      }
      else
        perror ("getpwuid()");
    }
#ifdef DEBUG
    printf ("running %s\n", fn);
#endif
    execl (fn, fn, NULL);
    exit (127);                 /* this will never execute */
  }
#if 0
  /* while we wait here the at trigger cannot be triggered */
  /* we have to catch sigchld another way */
  do
  {
    if (waitpid (pid, &status, 0) == -1)
    {
      if (errno != EINTR)
        return -1;
    }
    break;
  } while (1);
#endif
  return 0;
}



void
die (char *s)
{
  fprintf (stderr, "%s\n", s);
  exit (2);
}

void
die2 (char *s, char *a1)
{
  fprintf (stderr, s, a1);
  exit (2);
}

void
exit_atq (void)
{
}

void
alarmSignalHandler (int sig)
{
  struct dirent *dirent;
  unsigned long this, next, now;

  printf ("Alarm signal received! \n");
  if (signal (SIGALRM, alarmSignalHandler) == SIG_ERR)
    printf ("can't catch SIGALARM");
  printf ("register alarm signal again \n");

  if (sig == SIGALRM)
  {
    scan_from_top ();
    now = time (NULL);

    while ((dirent = scan ()))
    {
      this = nameok (dirent->d_name);
      printf ("this=%d\n", this);
      /* run jobs which are in the past, and find the next job in the future */
      /* avoid race conditions. Run jobs scheduled for the next second now */
      if (this)
        if (this <= now + 1)
        {
          char filename[100];

          runjob (dirent->d_name);

          sprintf (filename, "%s/%s", dirs[0], dirent->d_name);

          printf ("filename = %s\n", filename);

        }
    }                           /* end while */

    sleep (1);
  }
  return;
}

int
main (int argc, char *argv[])
{
  char *argv0;
  struct dirent *dirent;
  unsigned long this, now, next;

  argv0 = strrchr (argv[0], '/');
  if (!argv0)
    argv0 = argv[0];

  if (argc != 2)
    die2 ("usage: %s spooldir\n", argv0);

  if (chdir (argv[1]) < 0)
    die ("cannot chdir");
  if (!strcmp (argv0, "atq"))

    exit_atq ();
  if (mkfifo ("trigger.new", 0777) < 0)
    die ("cannot mkfifo trigger.new");
  if (rename ("trigger.new", "trigger"))
    die ("cannot rename trigger.new");
  chmod ("trigger", S_IWUSR | S_IWGRP | S_IWOTH);

  /* ignore some signals we get from spawned processes */
  signal (SIGCHLD, SIG_IGN);
  signal (SIGPIPE, SIG_IGN);

  if (signal (SIGALRM, alarmSignalHandler) == SIG_ERR)
    printf ("can't catch SIGALARM");

  /*
   * Run all the job in the past.
   */
  now = time (NULL);
#ifdef DEBUG
  printf ("%s:now=%d\n", __FUNCTION__, now);
#endif
  scan_from_top ();

  while ((dirent = scan ()))
  {
    this = nameok (dirent->d_name);
    printf ("%s:this=%d\n", __FUNCTION__, this);
    if (this)
      if (this <= now + 1)
        runjob (dirent->d_name);
  }


  while (1)
  {
    printf ("cilcle again\n");
/*
 * Look for the next alarm.
 */
    next = setNearestAlarm ();

#ifdef DEBUG
    printf ("%s:next: %ld\n", __FUNCTION__, next);
#endif
    waitfor (next);
  }
  return 0;
}

/*
vi:ts=2:nowrap:ai:expandtab
*/

